/*
 * BitVector.cpp
 *
 *  Created on: Feb 2, 2009
 *      Author: jonathanmasci
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdexcept>
#include "bitVector.h"

BitVector::BitVector(uint length) :
  _to_delete_buffer(true),
  _size(length)
  {
    _buffer_size = ceil(length / _BIT_WORD_SIZE);
    _buffer = (uchar *)malloc(_buffer_size * sizeof(*_buffer));
    memset(_buffer, 0, _buffer_size);
  }

BitVector::BitVector(uchar *buffer, uint length, uint bits) :
  _to_delete_buffer(true),
  _buffer(buffer),
  _buffer_size(length),
  _size(bits)
  {
  }

BitVector::~BitVector()
{
  if (_to_delete_buffer)
    delete[] _buffer;
}

/**
* Set the bit at the given position.
*
* @param pos the position of the bit to set
*/
void
BitVector::set(uint pos)
{
  uint byte = getByte(pos);
  uint bit = getBitInByte(pos, byte);

  if (checkRange(pos))
    _buffer[byte] |= 1 << bit;
  else
    {
      char c[15];
      sprintf(c, "%i", pos);
      throw std::out_of_range(std::string("Cannot access element ").append(c));
    }
}

void
BitVector::unset(uint pos)
{
  uint byte = getByte(pos);
  uint bit = getBitInByte(pos, byte);

  if (checkRange(pos))
    _buffer[byte] &= ~(1 << bit);
  else
    {
      char c[15];
      sprintf(c, "%i", pos);
      throw std::out_of_range(std::string("Cannot access element ").append(c));
    }
}

/**
* Get the value of the bit at the given position.
*
* @param pos the position of the bit
* @return bool true if bit is set
*/
bool BitVector::get(uint pos)
{
  uint byte = getByte(pos);
  uint bit = getBitInByte(pos, byte);

  if (checkRange(pos))
    return _buffer[byte] & 1 << bit;
  else
    {
      char c[15];
      sprintf(c, "%i", pos);
      throw std::out_of_range(std::string("Cannot access element ").append(c));
    }
}

/**
* The container where all bits are packed up.
*
* @return uchar * bits container
*/
uchar *BitVector::getBitContainer()
{
  return _buffer;
}

/**
* Invert the value of the bit at the given position.
*
* @param pos position of the bit to flip.
* @throw std::out_of_range
*/
void BitVector::flip(uint pos)
{
  uint byte = getByte(pos);
  uint bit = getBitInByte(pos, byte);

  if (checkRange(pos))
    _buffer[byte] = (_buffer[byte] ^ 1 << bit);
  else
    {
      char c[15];
      sprintf(c, "%i", pos);
      throw std::out_of_range(std::string("Cannot access element ").append(c));
    }
}

/**
* Check if the position is within the available range.
*
* @param pos the position to check
* @return true on success
*/
const bool
BitVector::checkRange(uint pos) const
{
  if (pos < _size && pos >= 0)
    return true;
  else
    return false;
}

/**
* Return a printable string of the BitVector.
*
* @return readable string representing all bits.
*/
uchar *
BitVector::toString()
{
  uint i;
  uchar *ret = new uchar[_size];
  for (i = 0; i < _size; ++i)
    if (get(i))
      ret[i] = '1';
    else
      ret[i] = '0';
  ret[_size] = '\0';
  return ret;
}

/**
* Return the byte containing the given bit.
*
* @param pos the position of the bit
* @return the byte that pack the bit
*/
const uint
BitVector::getByte(uint pos) const
{
  return pos / _BIT_WORD_SIZE;
}

/**
* Return the position of the bit within the containing byte.
*
* @param pos the position of the bit
* @param byte the containing byte obtained by getByte()
* @return bit position within byte
*/
const uint
BitVector::getBitInByte(uint pos, uint byte) const
{
  return pos - byte * _BIT_WORD_SIZE;
}

/**
* The number of bits.
*
* @return the size of the structure in bit.
*/
uint
BitVector::size()
{
  return _size;
}

/**
* Set all bits to 0.
*/
void
BitVector::clear()
{
  memset(_buffer, 0, _buffer_size);
}

/**
* Subscript operator to wrap a bit into a suitable structure that allow us
* to play with it like it was a bit reference.
*
* @param pos the bit we want to manipulate
* @return _bit_reference& that wrap up our bit
*/
BitVector::_bit_reference&
BitVector::operator[](uint pos)
{
  uint byte = getByte(pos);
  uint bit = getBitInByte(pos, byte);

  if (checkRange(pos))
    {
      _bit_reference *br = new _bit_reference(&_buffer[byte], 1 << bit);
      return *br;
    }
  else
    {
      char c[15];
      sprintf(c, "%i", pos);
      throw std::out_of_range(std::string("Cannot access element ").append(c));
    }
}

const BitVector::_bit_reference&
BitVector::operator[](uint pos) const
{
  uint byte = getByte(pos);
  uint bit = getBitInByte(pos, byte);

  if (checkRange(pos))
    {
      _bit_reference *br = new _bit_reference(&_buffer[byte], 1 << bit);
      return *br;
    }
  else
    {
      char c[15];
      sprintf(c, "%i", pos);
      throw std::out_of_range(std::string("Cannot access element ").append(c));
    }
}
