#include <cppunit/TestFixture.h>
#include <cppunit/TestCase.h>
#include <cppunit/Test.h>
#include <cppunit/TestCaller.h>
#include <cppunit/TestResult.h>
#include <cppunit/TestSuite.h>
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>
#include <stdexcept>
#include "bitVector.h"
#include "bitStream.h"

class BitVectorTest : public CppUnit::TestFixture
{
  // cppunit macro to avoid error prone test suite definition
  CPPUNIT_TEST_SUITE(BitVectorTest);
  CPPUNIT_TEST(subscriptSetAllTrueTest);
  CPPUNIT_TEST(subscriptSetAllFalseTest);
  CPPUNIT_TEST(subscriptSetSomeFalse);
  CPPUNIT_TEST(subscriptSetSomeTrue);
  CPPUNIT_TEST(setAllTrueTest);
  CPPUNIT_TEST(setAllFalseTest);
  CPPUNIT_TEST(flipTest);
  CPPUNIT_TEST(sizeTest);
  CPPUNIT_TEST(outOfBoundTest);
  CPPUNIT_TEST(subscriptReverseAllTrueTest);
  CPPUNIT_TEST(subscriptReverseAllFalseTest);
  CPPUNIT_TEST(onStackVectorTest);
  CPPUNIT_TEST(fromCharTest);
  CPPUNIT_TEST(toStringTest);
  CPPUNIT_TEST(clearTest);
  CPPUNIT_TEST_SUITE_END();

public:
  void
  setUp()
  {
    bvp = new BitVector(1000);
  }

  void
  tearDown()
  {
    delete bvp;
  }

  void
  sizeTest()
  {
    CPPUNIT_ASSERT(bvp->size() == 1000);
  }

  void
  outOfBoundTest()
  {
    CPPUNIT_ASSERT_THROW((*bvp)[100001] = true, std::out_of_range);
  }

  void
  clearTest()
  {
    unsigned int i;
    (*bvp)[10] = true;
    CPPUNIT_ASSERT((*bvp)[10] == true);
    bvp->clear();
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT((*bvp)[i] == false);
  }

  void
  subscriptSetAllTrueTest()
  {
    unsigned int i;
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT((*bvp)[i] == false);
    for (i = 0; i < bvp->size(); ++i)
      (*bvp)[i] = true;
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT((*bvp)[i] == true);
  }

  void
  subscriptSetAllFalseTest()
  {
    unsigned int i;
    for (i = 0; i < bvp->size(); ++i)
      (*bvp)[i] = true;
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT((*bvp)[i] == true);
    for (i = 0; i < bvp->size(); ++i)
      (*bvp)[i] = false;
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT((*bvp)[i] == false);
  }

  void
  subscriptSetSomeFalse()
  {
    unsigned int i;
    for (i = 0; i < bvp->size(); ++i)
      (*bvp)[i] = true;
    (*bvp)[8] = false;
    (*bvp)[147] = false;
    for (i = 0; i < bvp->size(); ++i)
      if (i == 8 || i == 147)
        CPPUNIT_ASSERT((*bvp)[i] == false);
      else
        CPPUNIT_ASSERT((*bvp)[i] == true);
  }

  void
  subscriptSetSomeTrue()
  {
    unsigned int i;
    for (i = 0; i < bvp->size(); ++i)
      (*bvp)[i] = false;
    (*bvp)[8] = true;
    (*bvp)[147] = true;
    for (i = 0; i < bvp->size(); ++i)
      if (i == 8 || i == 147)
        CPPUNIT_ASSERT((*bvp)[i] == true);
      else
        CPPUNIT_ASSERT((*bvp)[i] == false);
  }

  void
  setAllTrueTest()
  {
    unsigned int i;
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT(bvp->get(i) == false);
    for (i = 0; i < bvp->size(); ++i)
      bvp->set(i);
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT(bvp->get(i) == true);
  }

  void
  setAllFalseTest()
  {
    unsigned int i;
    for (i = 0; i < bvp->size(); ++i)
      bvp->set(i);
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT(bvp->get(i) == true);
    for (i = 0; i < bvp->size(); ++i)
      bvp->unset(i);
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT(bvp->get(i) == false);
  }

  void
  flipTest()
  {
    unsigned int i;
    for (i = 0; i < bvp->size(); ++i)
      bvp->set(i);
    for (i = 0; i < bvp->size(); ++i)
      bvp->flip(i);
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT(bvp->get(i) == false);
  }

  void
  subscriptReverseAllTrueTest()
  {
    unsigned int i;
    for (i = 0; i < bvp->size(); ++i)
      (*bvp)[i] = false;
    for (i = bvp->size() - 1; i < bvp->size(); --i)
      (*bvp)[i] = true;
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT((*bvp)[i] == true);
  }

  void
  subscriptReverseAllFalseTest()
  {
    unsigned int i;
    for (i = 0; i < bvp->size(); ++i)
      (*bvp)[i] = true;
    for (i = bvp->size() - 1; i < bvp->size(); --i)
      (*bvp)[i] = false;
    for (i = 0; i < bvp->size(); ++i)
      CPPUNIT_ASSERT((*bvp)[i] == false);
  }

  void
  onStackVectorTest()
  {
    BitVector sv(100000);
    unsigned int i;
    for (i = 0; i < sv.size(); ++i)
      CPPUNIT_ASSERT(sv[i] == false);
    for (i = 0; i < sv.size(); ++i)
      sv[i] = true;
    for (i = 0; i < sv.size(); ++i)
      CPPUNIT_ASSERT(sv[i] == true);
  }

  void
  fromCharTest()
  {
    unsigned int i;
    unsigned char *str = new unsigned char[100000];
    memset(str, 0, 100000);
    BitVector bv(str, 100000, 500000);
    for (i = 0; i < bv.size(); ++i)
      CPPUNIT_ASSERT(bv[i] == false);
    CPPUNIT_ASSERT_THROW(bv[500001] = false, std::out_of_range);
    for (i = 0; i < bv.size(); ++i)
      bv[i] = true;
    for (i = 0; i < bv.size(); ++i)
      CPPUNIT_ASSERT(bv[i] == true);
  }

  void
  toStringTest()
  {
    BitVector bv(5);
    bv[1] = true;
    bv[3] = true;
    CPPUNIT_ASSERT(strcmp((const char *)bv.toString(), "01010") == 0);
  }

private:
  BitVector *bvp;
};


class BitStreamTest : public CppUnit::TestFixture
{
  // cppunit macro to avoid error prone test suite definition
  CPPUNIT_TEST_SUITE(BitStreamTest);
  CPPUNIT_TEST(pushBackTest);
  CPPUNIT_TEST(toStringTest);
  CPPUNIT_TEST(readStreamTest);
  CPPUNIT_TEST_SUITE_END();

public:
  void
  setUp()
  {
    bsp = new BitStream();
  }

  void
  tearDown()
  {
    delete bsp;
  }

  void
  pushBackTest()
  {
    unsigned int i;
    for (i = 0; i < 1000; ++i)
      {
        bsp->push_back(true);
        CPPUNIT_ASSERT(bsp->size() == i + 1);
        CPPUNIT_ASSERT((*bsp)[i] == true);
      }
    CPPUNIT_ASSERT(bsp->size() == 1000);
    for (i = 0; i < 1000; ++i)
      CPPUNIT_ASSERT(bsp->get(i) == true);
  }

  void
  readStreamTest()
  {
    BitStream bs;
    unsigned int i = 0;
    bs.push_back(true);
    bs.push_back(false);
    bs.push_back(true);
    bs.push_back(true);
    bs.push_back(false);

    while (bs.can_read())
      {
        bs.read();
        ++i;
      }
    CPPUNIT_ASSERT(i == 5);
  }

  void
  toStringTest()
  {
    BitStream bs;
    bs.push_back(true);
    CPPUNIT_ASSERT(bs[0] == true);
    CPPUNIT_ASSERT_THROW(bs[1] = false, std::out_of_range);
    bs.push_back(true);
    CPPUNIT_ASSERT(bs[1] == true);
    CPPUNIT_ASSERT_THROW(bs[2] = false, std::out_of_range);
    bs.push_back(false);
    CPPUNIT_ASSERT(bs[2] == false);
    CPPUNIT_ASSERT_THROW(bs[3] = false, std::out_of_range);
    bs.push_back(false);
    CPPUNIT_ASSERT(bs[3] == false);
    CPPUNIT_ASSERT_THROW(bs[4] = false, std::out_of_range);

    CPPUNIT_ASSERT(bs.size() == 4);
    CPPUNIT_ASSERT(strcmp((const char *)bs.toString(), "1100") == 0);
  }

private:
  BitStream *bsp;
};

int main(int argc, char **argv)
{
  CppUnit::TextUi::TestRunner runner;
  runner.addTest(BitVectorTest::suite());
  runner.addTest(BitStreamTest::suite());
  runner.run();

  return 0;
}
